@@ -8,27 +8,21 @@ import time
 import traceback
 from functools import partial
 from multiprocessing.context import BaseContext
-from typing import (
-    Optional,
-    Any,
-    Dict,
-    Union,
-    List,
-    cast,
-    Sequence,
-)
+from typing import Any, Dict, List, Optional, Sequence, Union, cast
 
+import filelock
 import torch
 import torch.distributed as dist  # type: ignore
 import torch.distributions  # type: ignore
 import torch.multiprocessing as mp  # type: ignore
 import torch.nn as nn
 import torch.optim as optim
-
 # noinspection PyProtectedMember
 from torch._C._distributed_c10d import ReduceOp
 
-from allenact.algorithms.onpolicy_sync.misc import TrackingInfoType, TrackingInfo
+from allenact.algorithms.onpolicy_sync.misc import TrackingInfo, TrackingInfoType
+from allenact.base_abstractions.sensor import Sensor
+from allenact.utils.misc_utils import str2bool
 from allenact.utils.model_utils import md5_hash_of_state_dict
 
 try:
@@ -44,39 +38,51 @@ from allenact.algorithms.onpolicy_sync.policy import ActorCriticModel
 from allenact.algorithms.onpolicy_sync.storage import (
     ExperienceStorage,
     MiniBatchStorageMixin,
-    StreamingStorageMixin,
     RolloutStorage,
+    StreamingStorageMixin,
 )
 from allenact.algorithms.onpolicy_sync.vector_sampled_tasks import (
-    VectorSampledTasks,
+    COMPLETE_TASK_CALLBACK_KEY,
     COMPLETE_TASK_METRICS_KEY,
     SingleProcessVectorSampledTasks,
-)
-from allenact.base_abstractions.experiment_config import ExperimentConfig, MachineParams
-from allenact.base_abstractions.misc import (
-    RLStepResult,
-    Memory,
-    ActorCriticOutput,
-    GenericAbstractLoss,
+    VectorSampledTasks,
 )
 from allenact.base_abstractions.distributions import TeacherForcingDistr
+from allenact.base_abstractions.experiment_config import ExperimentConfig, MachineParams
+from allenact.base_abstractions.misc import (
+    ActorCriticOutput,
+    GenericAbstractLoss,
+    Memory,
+    RLStepResult,
+)
 from allenact.utils import spaces_utils as su
 from allenact.utils.experiment_utils import (
-    set_seed,
-    TrainingPipeline,
     LoggingPackage,
     PipelineStage,
-    set_deterministic_cudnn,
     ScalarMeanTracker,
     StageComponent,
+    TrainingPipeline,
+    set_deterministic_cudnn,
+    set_seed,
 )
 from allenact.utils.system import get_logger
-from allenact.utils.tensor_utils import (
-    batch_observations,
-    detach_recursively,
-)
+from allenact.utils.tensor_utils import batch_observations, detach_recursively
 from allenact.utils.viz_utils import VizSuite
 
+try:
+    # When debugging we don't want to timeout in the VectorSampledTasks
+
+    # noinspection PyPackageRequirements
+    import pydevd
+
+    DEBUGGING = str2bool(os.getenv("ALLENACT_DEBUG", "true"))
+except ImportError:
+    DEBUGGING = str2bool(os.getenv("ALLENACT_DEBUG", "false"))
+
+DEBUG_VST_TIMEOUT: Optional[int] = (lambda x: int(x) if x is not None else x)(
+    os.getenv("ALLENACT_DEBUG_VST_TIMEOUT", None)
+)
+
 TRAIN_MODE_STR = "train"
 VALID_MODE_STR = "valid"
 TEST_MODE_STR = "test"
@@ -102,6 +108,7 @@ class OnPolicyRLEngine(object):
         ],  # to write/read (trainer/evaluator) ready checkpoints
         checkpoints_dir: str,
         mode: str = "train",
+        callback_sensors: Optional[Sequence[Sensor]] = None,
         seed: Optional[int] = None,
         deterministic_cudnn: bool = False,
         mp_ctx: Optional[BaseContext] = None,
@@ -113,6 +120,7 @@ class OnPolicyRLEngine(object):
         deterministic_agents: bool = False,
         max_sampler_processes_per_worker: Optional[int] = None,
         initial_model_state_dict: Optional[Union[Dict[str, Any], int]] = None,
+        try_restart_after_task_error: bool = False,
         **kwargs,
     ):
         """Initializer.
@@ -140,6 +148,7 @@ class OnPolicyRLEngine(object):
         self.device = torch.device("cpu") if device == -1 else torch.device(device)  # type: ignore
         self.distributed_ip = distributed_ip
         self.distributed_port = distributed_port
+        self.try_restart_after_task_error = try_restart_after_task_error
 
         self.mode = mode.lower().strip()
         assert self.mode in [
@@ -148,6 +157,7 @@ class OnPolicyRLEngine(object):
             TEST_MODE_STR,
         ], 'Only "train", "valid", "test" modes supported'
 
+        self.callback_sensors = callback_sensors
         self.deterministic_cudnn = deterministic_cudnn
         if self.deterministic_cudnn:
             set_deterministic_cudnn()
@@ -230,6 +240,9 @@ class OnPolicyRLEngine(object):
                 port=self.distributed_port,
                 world_size=self.num_workers,
                 is_master=self.worker_id == 0,
+                timeout=datetime.timedelta(
+                    seconds=3 * (DEBUG_VST_TIMEOUT if DEBUGGING else 1 * 60) + 300
+                ),
             )
             cpu_device = self.device == torch.device("cpu")  # type:ignore
 
@@ -246,7 +259,7 @@ class OnPolicyRLEngine(object):
                 # During testing, we sometimes found that default timeout was too short
                 # resulting in the run terminating surprisingly, we increase it here.
                 timeout=datetime.timedelta(minutes=3000)
-                if self.mode == TEST_MODE_STR
+                if (self.mode == TEST_MODE_STR or DEBUGGING)
                 else dist.default_pg_timeout,
             )
             self.is_distributed = True
@@ -262,6 +275,7 @@ class OnPolicyRLEngine(object):
 
         # Keeping track of metrics during training/inference
         self.single_process_metrics: List = []
+        self.single_process_task_callback_data: List = []
 
     @property
     def vector_tasks(
@@ -295,11 +309,13 @@ class OnPolicyRLEngine(object):
             self._vector_tasks = VectorSampledTasks(
                 make_sampler_fn=self.config.make_sampler_fn,
                 sampler_fn_args=self.get_sampler_fn_args(seeds),
+                callback_sensors=self.callback_sensors,
                 multiprocessing_start_method="forkserver"
                 if self.mp_ctx is None
                 else None,
                 mp_ctx=self.mp_ctx,
                 max_processes=self.max_sampler_processes_per_worker,
+                read_timeout=DEBUG_VST_TIMEOUT if DEBUGGING else 1 * 60,
             )
         return self._vector_tasks
 
@@ -572,14 +588,17 @@ class OnPolicyRLEngine(object):
 
         # Save after task completion metrics
         for step_result in outputs:
-            if (
-                step_result.info is not None
-                and COMPLETE_TASK_METRICS_KEY in step_result.info
-            ):
-                self.single_process_metrics.append(
-                    step_result.info[COMPLETE_TASK_METRICS_KEY]
-                )
-                del step_result.info[COMPLETE_TASK_METRICS_KEY]
+            if step_result.info is not None:
+                if COMPLETE_TASK_METRICS_KEY in step_result.info:
+                    self.single_process_metrics.append(
+                        step_result.info[COMPLETE_TASK_METRICS_KEY]
+                    )
+                    del step_result.info[COMPLETE_TASK_METRICS_KEY]
+                if COMPLETE_TASK_CALLBACK_KEY in step_result.info:
+                    self.single_process_task_callback_data.append(
+                        step_result.info[COMPLETE_TASK_CALLBACK_KEY]
+                    )
+                    del step_result.info[COMPLETE_TASK_CALLBACK_KEY]
 
         rewards: Union[List, torch.Tensor]
         observations, rewards, dones, infos = [list(x) for x in zip(*outputs)]
@@ -840,6 +859,36 @@ class OnPolicyTrainer(OnPolicyRLEngine):
             )  # use the latest seed for workers and update rng state
             self.vector_tasks.set_seeds(seeds)
 
+    def save_error_data(self, batch: Dict[str, Any]) -> str:
+        model_path = os.path.join(
+            self.checkpoints_dir,
+            "error_for_exp_{}__stage_{:02d}__steps_{:012d}.pt".format(
+                self.experiment_name,
+                self.training_pipeline.current_stage_index,
+                self.training_pipeline.total_steps,
+            ),
+        )
+        with filelock.FileLock(
+            os.path.join(self.checkpoints_dir, "error.lock"), timeout=60
+        ):
+            if not os.path.exists(model_path):
+                save_dict = {
+                    "model_state_dict": self.actor_critic.state_dict(),  # type:ignore
+                    "total_steps": self.training_pipeline.total_steps,  # Total steps including current stage
+                    "optimizer_state_dict": self.optimizer.state_dict(),  # type: ignore
+                    "training_pipeline_state_dict": self.training_pipeline.state_dict(),
+                    "trainer_seed": self.seed,
+                    "batch": batch,
+                }
+
+                if self.lr_scheduler is not None:
+                    save_dict["scheduler_state"] = cast(
+                        _LRScheduler, self.lr_scheduler
+                    ).state_dict()
+
+                torch.save(save_dict, model_path)
+        return model_path
+
     def checkpoint_save(self, pipeline_stage_index: Optional[int] = None) -> str:
         model_path = os.path.join(
             self.checkpoints_dir,
@@ -1108,12 +1157,21 @@ class OnPolicyTrainer(OnPolicyRLEngine):
                         bsize = batch["bsize"]
 
                         if actor_critic_output_for_batch is None:
-                            actor_critic_output_for_batch, _ = self.actor_critic(
-                                observations=batch["observations"],
-                                memory=batch["memory"],
-                                prev_actions=batch["prev_actions"],
-                                masks=batch["masks"],
-                            )
+
+                            try:
+                                actor_critic_output_for_batch, _ = self.actor_critic(
+                                    observations=batch["observations"],
+                                    memory=batch["memory"],
+                                    prev_actions=batch["prev_actions"],
+                                    masks=batch["masks"],
+                                )
+                            except ValueError:
+                                save_path = self.save_error_data(batch=batch)
+                                get_logger().error(
+                                    f"Encountered a value error! Likely because of nans in the output/input."
+                                    f" Saving all error information to {save_path}."
+                                )
+                                raise
 
                         loss_return = loss.loss(
                             step_count=self.step_count,
@@ -1271,6 +1329,10 @@ class OnPolicyTrainer(OnPolicyRLEngine):
 
         self.aggregate_task_metrics(logging_pkg=logging_pkg)
 
+        for callback_dict in self.single_process_task_callback_data:
+            logging_pkg.task_callback_data.append(callback_dict)
+        self.single_process_task_callback_data = []
+
         if self.mode == TRAIN_MODE_STR:
             # Technically self.mode should always be "train" here (as this is the training engine),
             # this conditional is defensive
@@ -1312,7 +1374,7 @@ class OnPolicyTrainer(OnPolicyRLEngine):
                 self.checkpoints_queue.put(("eval", model_path))
         self.last_save = self.training_pipeline.total_steps
 
-    def run_pipeline(self):
+    def run_pipeline(self, valid_on_initial_weights: bool = False):
         cur_stage_training_settings = (
             self.training_pipeline.current_stage.training_settings
         )
@@ -1336,6 +1398,16 @@ class OnPolicyTrainer(OnPolicyRLEngine):
         )
         already_saved_checkpoint = False
 
+        if (
+            valid_on_initial_weights
+            and should_save_checkpoints
+            and self.checkpoints_queue is not None
+        ):
+            if self.worker_id == self.first_local_worker_id:
+                model_path = self.checkpoint_save()
+                if self.checkpoints_queue is not None:
+                    self.checkpoints_queue.put(("eval", model_path))
+
         while True:
             pipeline_stage_changed = self.training_pipeline.before_rollout(
                 train_metrics=self._last_aggregated_train_task_metrics
@@ -1414,11 +1486,49 @@ class OnPolicyTrainer(OnPolicyRLEngine):
                 for k, v in self.training_pipeline.current_stage_storage.items()
             }
 
-            for step in range(cur_stage_training_settings.num_steps):
-                num_paused = self.collect_step_across_all_task_samplers(
-                    rollout_storage_uuid=self.training_pipeline.rollout_storage_uuid,
-                    uuid_to_storage=uuid_to_storage,
-                )
+            vector_tasks_already_restarted = False
+            step = -1
+            while step < cur_stage_training_settings.num_steps - 1:
+                step += 1
+
+                try:
+                    num_paused = self.collect_step_across_all_task_samplers(
+                        rollout_storage_uuid=self.training_pipeline.rollout_storage_uuid,
+                        uuid_to_storage=uuid_to_storage,
+                    )
+                except (TimeoutError, EOFError) as e:
+                    if (
+                        not self.try_restart_after_task_error
+                    ) or self.mode != TRAIN_MODE_STR:
+                        # Apparently you can just call `raise` here and doing so will just raise the exception as though
+                        # it was not caught (so the stacktrace isn't messed up)
+                        raise
+                    elif vector_tasks_already_restarted:
+                        raise RuntimeError(
+                            f"[{self.mode} worker {self.worker_id}] `vector_tasks` has timed out twice in the same"
+                            f" rollout. This suggests that this error was not recoverable. Timeout exception:\n{traceback.format_exc()}"
+                        )
+                    else:
+                        get_logger().warning(
+                            f"[{self.mode} worker {self.worker_id}] `vector_tasks` appears to have crashed during"
+                            f" training due to an {type(e).__name__} error. You have set"
+                            f" `try_restart_after_task_error` to `True` so we will attempt to restart these tasks from"
+                            f" the beginning. USE THIS FEATURE AT YOUR OWN"
+                            f" RISK. Exception:\n{traceback.format_exc()}."
+                        )
+                        self.vector_tasks.close()
+                        self._vector_tasks = None
+
+                        vector_tasks_already_restarted = True
+                        for (
+                            storage
+                        ) in self.training_pipeline.current_stage_storage.values():
+                            storage.after_updates()
+                        self.initialize_storage_and_viz(
+                            storage_to_initialize=list(uuid_to_storage.values())
+                        )
+                        step = -1
+                        continue
 
                 # A more informative error message should already have been thrown in be given in
                 # `collect_step_across_all_task_samplers` if `num_paused != 0` here but this serves
@@ -1572,7 +1682,10 @@ class OnPolicyTrainer(OnPolicyRLEngine):
                 )
 
     def train(
-        self, checkpoint_file_name: Optional[str] = None, restart_pipeline: bool = False
+        self,
+        checkpoint_file_name: Optional[str] = None,
+        restart_pipeline: bool = False,
+        valid_on_initial_weights: bool = False,
     ):
         assert (
             self.mode == TRAIN_MODE_STR
@@ -1584,7 +1697,7 @@ class OnPolicyTrainer(OnPolicyRLEngine):
             if checkpoint_file_name is not None:
                 self.checkpoint_load(checkpoint_file_name, restart_pipeline)
 
-            self.run_pipeline()
+            self.run_pipeline(valid_on_initial_weights=valid_on_initial_weights)
 
             training_completed_successfully = True
         except KeyboardInterrupt:
@@ -1595,7 +1708,7 @@ class OnPolicyTrainer(OnPolicyRLEngine):
             get_logger().error(
                 f"[{self.mode} worker {self.worker_id}] Encountered {type(e).__name__}, exiting."
             )
-            get_logger().exception(traceback.format_exc())
+            get_logger().error(traceback.format_exc())
         finally:
             if training_completed_successfully:
                 if self.worker_id == 0:
@@ -1786,6 +1899,10 @@ class OnPolicyInference(OnPolicyRLEngine):
 
         self.aggregate_task_metrics(logging_pkg=logging_pkg)
 
+        for callback_dict in self.single_process_task_callback_data:
+            logging_pkg.task_callback_data.append(callback_dict)
+        self.single_process_task_callback_data = []
+
         logging_pkg.viz_data = (
             visualizer.read_and_reset() if visualizer is not None else None
         )
