@@ -19,8 +19,8 @@ IMAGENET_RGB_STDS: Tuple[float, float, float] = (0.229, 0.224, 0.225)
 class VisionSensor(Sensor[EnvType, SubTaskType]):
     def __init__(
         self,
-        mean: Optional[np.ndarray] = None,
-        stdev: Optional[np.ndarray] = None,
+        mean: Union[Sequence[float], np.ndarray, None] = None,
+        stdev: Union[Sequence[float], np.ndarray, None] = None,
         height: Optional[int] = None,
         width: Optional[int] = None,
         uuid: str = "vision",
@@ -50,8 +50,8 @@ class VisionSensor(Sensor[EnvType, SubTaskType]):
         kwargs : Extra kwargs. Currently unused.
         """
 
-        self._norm_means = mean
-        self._norm_sds = stdev
+        self._norm_means = np.array(mean)
+        self._norm_sds = np.array(stdev)
         assert (self._norm_means is None) == (self._norm_sds is None), (
             "In VisionSensor's config, "
             "either both mean/stdev must be None or neither."
@@ -162,28 +162,37 @@ class VisionSensor(Sensor[EnvType, SubTaskType]):
 
     def process_img(self, img: np.ndarray):
         assert (
-            img.dtype == np.float32 and (len(img.shape) == 2 or img.shape[-1] == 1)
-        ) or (img.shape[-1] == 3 and img.dtype == np.uint8), (
+            np.issubdtype(img.dtype, np.float32)
+            and (len(img.shape) == 2 or img.shape[-1] == 1)
+        ) or (img.shape[-1] == 3 and np.issubdtype(img.dtype, np.uint8)), (
             "Input frame must either have 3 channels and be of"
             " type np.uint8 or have one channel and be of type np.float32"
         )
 
-        if self._scale_first:
-            if self.scaler is not None and img.shape[:2] != (self._height, self._width):
-                img = np.array(self.scaler(self.to_pil(img)), dtype=img.dtype)  # hwc
+        if (
+            self._scale_first
+            and self.scaler is not None
+            and img.shape[:2] != (self._height, self._width)
+        ):
+            img = np.array(self.scaler(self.to_pil(img)), dtype=img.dtype)  # hwc
+        elif np.issubdtype(img.dtype, np.float32):
+            img = img.copy()
 
         assert img.dtype in [np.uint8, np.float32]
 
-        if img.dtype == np.uint8:
+        if np.issubdtype(img.dtype, np.uint8):
             img = img.astype(np.float32) / 255.0
 
         if self._should_normalize:
             img -= self._norm_means
             img /= self._norm_sds
 
-        if not self._scale_first:
-            if self.scaler is not None and img.shape[:2] != (self._height, self._width):
-                img = np.array(self.scaler(self.to_pil(img)), dtype=np.float32)  # hwc
+        if (
+            (not self._scale_first)
+            and self.scaler is not None
+            and img.shape[:2] != (self._height, self._width)
+        ):
+            img = np.array(self.scaler(self.to_pil(img)), dtype=np.float32)  # hwc
 
         return img
 
